"""
Convention-driven pubsub.

"""
from enum import Enum, unique
from inspect import isclass
from six import string_types

from inflection import underscore
from marshmallow import fields

from microcosm_pubsub.codecs import PubSubMessageSchema


@unique
class LifecycleChange(Enum):
    """
    CRUD lifecycle changes that may be announced via pubsub.

    Note that `Changed` is strongly discouraged; use immutable resources instead.

    """
    Changed = u"changed"
    Created = u"created"
    Deleted = u"deleted"

    def matches(self, media_type):
        return self.value in media_type.split(".")


def name_for(obj):
    """
    Get a name for something.

    Allows overriding of default names using the `__alias__` attribute.

    """
    if isinstance(obj, string_types):
        return underscore(obj)

    cls = obj if isclass(obj) else obj.__class__

    if hasattr(cls, "__alias__"):
        return underscore(cls.__alias__)
    else:
        return underscore(cls.__name__)


def make_media_type(resource, lifecycle_change=None, organization=None, public=False):
    """
    Generate a canonical media type for a message that announces a lifecycle change to a resource.

    :param resource: the resource (name)
    :param lifecycle_change: the change that happened to the resource
    :param organization: the vendor organization to which this resource belongs
    :param public: whether the resource is meant for public/external consumption

    """
    if lifecycle_change is None:
        lifecycle_change = LifecycleChange.Created
    if organization is None:
        organization = "globality"

    return "application/vnd.{}.{}.{}.{}".format(
        # use a vendor specific media type
        organization,
        # differentiate messages sent over a pubsub channel from other messages
        "pubsub" if public else "pubsub._",
        # qualify the message with the kind of lifecycle change
        lifecycle_change.value,
        # specify the resource name
        name_for(resource),
    )


class URIMessageSchema(PubSubMessageSchema):
    """
    Define a baseline message schema that points to (the URI of) a resource that experienced a lifecycle change.

    By convention, pubsub messages should be *references* to something that happened within some other
    source-of-truth (e.g. a CRUD microservice). Because there are inherent race conditions between publishing
    a message and commit a persistent transaction, message consumers are expected to call back to source-of-truth
    (e.g. via HTTP) and fetch the current resource value. In the event that message has not been committed yet,
    the consumer can retry a few times before giving up (e.g. via SQS dead-lettering).

    """
    def __init__(self, media_type, **kwargs):
        super(PubSubMessageSchema, self).__init__(**kwargs)
        self.MEDIA_TYPE = media_type

    uri = fields.String(required=True)

    def deserialize_media_type(self, obj):
        return self.MEDIA_TYPE


def changed(resource, **kwargs):
    """
    Fluent wrapper around message publishing for convention-driven schemas.

    """
    return make_media_type(resource, lifecycle_change=LifecycleChange.Changed, **kwargs)


def created(resource, **kwargs):
    """
    Fluent wrapper around message publishing for convention-driven schemas.

    """
    return make_media_type(resource, lifecycle_change=LifecycleChange.Created, **kwargs)


def deleted(resource, **kwargs):
    """
    Fluent wrapper around message publishing for convention-driven schemas.

    """
    return make_media_type(resource, lifecycle_change=LifecycleChange.Deleted, **kwargs)



"""
Consume Daemon main.

"""
import microcosm.opaque  # noqa
from microcosm_daemon.api import SleepNow
from microcosm_daemon.daemon import Daemon


class ConsumerDaemon(Daemon):

    def make_arg_parser(self):
        parser = super(ConsumerDaemon, self).make_arg_parser()
        parser.add_argument("--sqs-queue-url")
        return parser

    def create_object_graph_components(self, graph):
        super(ConsumerDaemon, self).create_object_graph_components(graph)

        # legacy schema mappings support (as property)
        if hasattr(self, "schema_mappings"):
            for media_type, schema_cls in self.schema_mappings.items():
                graph.pubsub_message_schema_registry.register(media_type, schema_cls)

    def run_state_machine(self):
        for media_type in self.graph.sqs_message_handler_registry.keys():
            handler = self.graph.sqs_message_handler_registry[media_type]
            self.graph.logger.info("Handling: {} with handler: {}".format(
                media_type,
                handler.__class__.__name__,
            ))
        super(ConsumerDaemon, self).run_state_machine()

    @property
    def defaults(self):
        if not self.args.sqs_queue_url:
            return dict()

        return dict(
            sqs_consumer=dict(
                sqs_queue_url=self.args.sqs_queue_url,
            ),
        )

    @property
    def components(self):
        return super(ConsumerDaemon, self).components + [
            "opaque",
            "pubsub_message_schema_registry",
            "sqs_message_handler_registry",
            "sqs_consumer",
            "sqs_message_dispatcher",
        ]

    def __call__(self, graph):
        """
        Implement daemon by sinking messages from the consumer to a dispatcher function.

        """
        result = graph.sqs_message_dispatcher.handle_batch()
        if not result.message_count:
            raise SleepNow()



"""
Message producer.

"""
from collections import defaultdict

from boto3 import client
from microcosm.api import defaults
from microcosm.errors import NotBoundError
from microcosm_logging.decorators import logger
from microcosm_logging.timing import elapsed_time

from microcosm_pubsub.errors import TopicNotDefinedError


@logger
class SNSProducer(object):
    """
    Produces messages to SNS topics.

    """
    def __init__(self, opaque, pubsub_message_schema_registry, sns_client, sns_topic_arns):
        self.opaque = opaque
        self.pubsub_message_schema_registry = pubsub_message_schema_registry
        self.sns_client = sns_client
        self.sns_topic_arns = sns_topic_arns

    def produce(self, media_type, dct=None, **kwargs):
        """
        Produce a message.

        :returns: the message id

        """
        message, topic_arn, opaque_data = self.create_message(media_type, dct, **kwargs)
        return self.publish_message(media_type, message, topic_arn, opaque_data)

    def create_message(self, media_type, dct, opaque_data=None, **kwargs):
        if opaque_data is None:
            opaque_data = dict()

        if self.opaque is not None:
            opaque_data.update(self.opaque.as_dict())

        topic_arn = self.choose_topic_arn(media_type)
        message = self.pubsub_message_schema_registry[media_type].encode(
            dct,
            opaque_data=opaque_data,
            **kwargs
        )
        return message, topic_arn, opaque_data

    def publish_message(self, media_type, message, topic_arn, opaque_data):
        extra = dict(
            media_type=media_type,
            **opaque_data
        )
        self.logger.debug("Publishing message with media type {media_type}", extra=extra)

        with elapsed_time(extra):
            result = self.sns_client.publish(
                TopicArn=topic_arn,
                Message=message,
            )

        self.logger.info("Published message with media type {media_type}", extra)

        return result["MessageId"]

    def choose_topic_arn(self, media_type):
        """
        Choose a topic for this type of message.

        """
        try:
            topic_arn = self.sns_topic_arns[media_type]
        except KeyError:
            topic_arn = None

        if topic_arn is None:
            raise TopicNotDefinedError("No topic arn was registered for messages of type: {}".format(
                media_type,
            ))
        return topic_arn


class DeferredProducer(object):
    """
    A context manager to defer message production until the end of a block.

    """
    def __init__(self, producer):
        self.producer = producer
        self.messages = []

    def produce(self, media_type, dct=None, **kwargs):
        message, topic_arn, opaque_data = self.producer.create_message(media_type, dct, **kwargs)
        self.messages.append((media_type, message, topic_arn, opaque_data))

    def __enter__(self):
        self.message = []
        return self

    def __exit__(self, type, value, traceback):
        if type is not None:
            return

        for media_type, message, topic_arn, opaque_data in self.messages:
            self.producer.publish_message(media_type, message, topic_arn, opaque_data)


@defaults(
    default=None,
    mappings={},
)
def configure_sns_topic_arns(graph):
    """
    Configure a mapping from message types to topic ARNs.

    """
    if graph.config.sns_topic_arns.default is None:
        sns_topic_arns = dict()
    else:
        sns_topic_arns = defaultdict(lambda: graph.config.sns_topic_arns.default)

    sns_topic_arns.update(graph.config.sns_topic_arns.mappings)

    return sns_topic_arns


def configure_sns_producer(graph):
    """
    Configure an SNS producer.

    The SNS Producer requires the following collaborators:
        - Opaque from microcosm.opaque for capturing context information
        - an aws sns client, i.e. from boto.
        - pubsub message codecs: see tests for examples.
        - sns topic arns: see tests for examples.

    """
    if graph.metadata.testing:
        from mock import MagicMock
        sns_client = MagicMock()
    else:
        sns_client = client("sns")

    try:
        opaque = graph.opaque
    except NotBoundError:
        opaque = None

    return SNSProducer(
        opaque=opaque,
        pubsub_message_schema_registry=graph.pubsub_message_schema_registry,
        sns_client=sns_client,
        sns_topic_arns=graph.sns_topic_arns,
    )



"""
Parse the SQS envelope.

SQS defines an envelope with a message id, receipt handle, body, and so forth,
with the underlying message embedded in the body. There are multiple ways to
process this envelope, depending on the degree of validation and metadata desired.

"""
from abc import ABCMeta, abstractmethod
from json import loads
from hashlib import md5

from microcosm.api import defaults

from microcosm_pubsub.codecs import MediaTypeSchema, PubSubMessageCodec
from microcosm_pubsub.message import SQSMessage


class SQSEnvelope(object):
    """
    Enveloping base class.

    """
    __metaclass__ = ABCMeta

    def __init__(self, graph):
        self.validate_md5 = graph.config.sqs_envelope.validate_md5

    def parse_raw_message(self, consumer, raw_message):
        """
        Create an `SQSMessage` from SQS data.

        """
        message_id = raw_message["MessageId"]
        receipt_handle = raw_message["ReceiptHandle"]
        body = raw_message["Body"]

        if self.validate_md5:
            self.validate_md5(raw_message, body)

        message = loads(body)["Message"]
        media_type, content = self.parse_media_type_and_content(message)

        return SQSMessage(
            consumer=consumer,
            content=content,
            media_type=media_type,
            message_id=message_id,
            receipt_handle=receipt_handle,
        )

    @abstractmethod
    def parse_media_type_and_content(self, message):
        """
        Parse and validate SQS message media type and content.

        :returns: a media_type, content tuple

        """
        pass

    def validate_md5(self, raw_message, body):
        """
        Validate the message body.

        Just checks for tampering; schema validation occurs once we know the type of message.

        """
        expected_md5_of_body = raw_message["MD5OfBody"]
        actual_md5_of_body = md5(body).hexdigest()
        if expected_md5_of_body != actual_md5_of_body:
            raise Exception("MD5 validation failed. Expected: {} Actual: {}".format(
                expected_md5_of_body,
                actual_md5_of_body,
            ))


class RawSQSEnvelope(SQSEnvelope):
    """
    Enveloping strategy that just passes raw JSON.

    """
    def parse_media_type_and_content(self, message):
        content = loads(message)
        media_type = "application/json"
        return media_type, content


class CodecSQSEnvelope(SQSEnvelope):
    """
    Enveloping strategy that uses a media type-driven message codec.

    """
    def __init__(self, graph):
        super(CodecSQSEnvelope, self).__init__(graph)
        self.media_type_codec = PubSubMessageCodec(MediaTypeSchema())
        self.pubsub_message_schema_registry = graph.pubsub_message_schema_registry

    def parse_media_type_and_content(self, message):
        """
        Decode the message once to extract its media type and then again with the correct codec.

        """
        base_message = self.media_type_codec.decode(message)
        media_type = base_message["mediaType"]
        try:
            content = self.pubsub_message_schema_registry[media_type].decode(message)
        except KeyError:
            return media_type, None
        else:
            return media_type, content


@defaults(
    strategy_name="CodecSQSEnvelope",
    validate_md5=False,
)
def configure_sqs_envelope(graph):
    strategy_name = graph.config.sqs_envelope.strategy_name
    # It should be possible to inject custom enveloping strategies as long as the
    # subclass is imported before the graph is initialized; alternative, this block
    # can be switched to using setuptools entry points.
    strategies = {
        strategy.__name__: strategy
        for strategy in SQSEnvelope.__subclasses__()
    }
    try:
        strategy = strategies[strategy_name]
    except KeyError:
        raise Exception("Unknown SQS enveloping strategy: {}".format(
            strategy_name,
        ))
    else:
        return strategy(graph)
